<?php

/* vim: set expandtab tabstop=4 shiftwidth=4 softtabstop=4: */

/**
 * File_Fortune: Read fortune databases.
 *
 * Read fortune databases.
 *
 * If you wish to write to fortune databases, use {@link File_Fortune_Writer}.
 *
 * PHP version 5
 *
 * LICENSE: This source file is subject to version 3.0 of the PHP license
 * that is available through the world-wide-web at the following URI:
 * http://www.php.net/license/3_0.txt.  If you did not receive a copy of
 * the PHP License and are unable to obtain it through the web, please
 * send a note to license@php.net so we can mail you a copy immediately.
 *
 * @category  File_Formats
 * @package   File_Fortune
 * @author    Matthew Weier O'Phinney <mweierophinney@gmail.com>, based on
 *            Fortune.pm, by Greg Ward
 * @copyright 2005, Matthew Weier O'Phinney; Fortune.pm v0.2(c) 1999, Greg Ward
 * @version   CVS: $Id:$
 */

/**
 * Utilize exceptions for error handling
 */
require_once 'File/Fortune/Exception.php';

/**
 * File_Fortune: Interface to fortune cookie databases
 *
 * The <em>fortune</em> program is a small but important part of *nix
 * culture, and this package aims to provide support for its "fortune cookie"
 * databases to PHP programmers.  
 *
 * If you wish to be able to write or manipulate fortunes, you may be interested
 * in {@link File_Fortune_Writer}.
 *
 * @package   File_Fortune
 * @version   Release: 0.9.0
 * @tutorial  File_Fortune/File_Fortune.cls
 */
class File_Fortune
{
    /**
     * Version
     * @staticvar string 
     * @access public
     */
    static $VERSION = '0.9.0';

    /**
     * Character that delimits fortunes (defaults to '%')
     * @var string
     * @access public
     */
    public $delim;

    /**
     * Fortune file filehandle
     * @var resource 
     * @access public
     */
    protected $file;

    /**
     * Name of fortune file to use
     * @var string 
     * @access public
     */
    public $filename;

    /**
     * Bit field for flags
     * @var array
     * @access public
     */
    protected $flags;

    /**
     * Array of fortunes; unset by default.
     * @var array 
     * @access public
     */
    public $fortunes;

    /**
     * Name of fortune header file (.dat file)
     * @var string 
     * @access public
     */
    public $headerFilename;

    /**
     * Length of longest string in the fortune file
     * @var int
     * @access public
     */
    public $maxLength;

    /**
     * Length of shortest string in the fortune file
     * @var int
     * @access public
     */
    public $minLength;

    /**
     * Flag indicating whether or not a header file is present. Defaults to
     * false.
     * @var bool 
     * @access protected
     */
    protected $noHeader = false;

    /**
     * Number of strings (fortunes) in the file
     * @var int
     * @access public
     */
    public $numstr;

    /**
     * Array of file offsets (fortune indices)
     * @var array 
     * @access public
     */
    protected $offsets;

    /**
     * Version number for fortune file
     * @var int
     * @access public
     */
    public $version;

    /**
     * Constructor; initialize a fortune cookie database
     *
     * Initializes the {@link PEAR_ErrorStack error stack} and assigns it to
     * {@link $error}.
     *
     * If a fortune file is provided to the constructor, it attempts to open it.
     * If no $headerFilename is provided, "$filename.dat" will be used. The
     * open operation is performed by {@link open()}; if an error occurs with
     * that operation, an error will be pushed onto the 
     * {@link $error error stack}.
     *
     * If no data file is given, the class will be instantiated with no data,
     * and you will need to {@link open() open} a file before performing any
     * operations.
     *
     * @access public
     * @param string $filename Fortune file name
     * @param string $headerFilename Fortune header file name; defaults to
     * "$filename.dat"
     * @throws File_Fortune_Exception if unable to open file; see {@link open()}
     */
    public function __construct($filename = null, $headerFilename = null)
    {
        if (!empty($filename)) {
            try {
                $this->open($filename, $headerFilename);
            } catch (File_Fortune_Exception $e) {
                throw $e;
            }
        }
    }

    /**
     * Destructor
     *
     * Closes fortune file, utilizing {@link close()}.
     *
     * @access public
     */
    public function __destruct()
    {
        $this->close();
    }

    /**
     * Open a fortune file
     *
     * Opens a fortune file. If no $headerFilename is given, "$filename.dat" is
     * used; if $headerFilename is invalid or not present, the class will parse
     * the fortune file in order to build the necessary header elements.
     *
     * Once the file has been opened, its file handle will be assigned to 
     * {@link $file}.  Additionally, a number of other properties will be set:
     * <ul>
     *     <li>{@link $numstr}</li>
     *     <li>{@link $offsets}</li>
     *     <li>{@link $delim}</li>
     *     <li>{@link $flags}</li>
     *     <li>{@link $maxLength}</li>
     *     <li>{@link $minLength}</li>
     *     <li>{@link $version}</li>
     * </ul>
     *
     * If no header file is present, open() will read the entire file into
     * memory, and store all fortunes in {@link $fortunes}.
     *
     * Returns true on success; on failure, throws an exception.
     *
     * <b>Note:</b> If a file is already open when open() is called, that file
     * is first closed.
     * 
     * @access public
     * @param string $filename Name of fortune file (defaults to 
     * {@link $filename})
     * @param string $headerFilename Name of fortune file header file (defaults
     * to either {@link $headerFilename} or {@link $filename}.dat)
     * @return true
     * @throws File_Fortune_Exception_NoFilename if no filename passed and no
     * filename in object
     * @throws File_Fortune_Exception_BadFile if missing file or unable to read
     * from file
     * @throws File_Fortune_Exception_BadHeaderFile if unable to open header
     * file, unable to read header file, unable to get header information, or if
     * header information appears corrupted
     */
    public function open($filename = null, $headerFilename = null)
    {
        // First check for open file; close if so
        if (isset($this->file)) {
            $this->close();
        }

        // Get filename
        if (empty($filename) && empty($this->filename)) {
            // No filename; cannot open
            throw new File_Fortune_Exception_NoFilename('Missing filename');
        } elseif (!empty($filename)) {
            $this->filename = $filename;
        }

        // Get header filename
        if (empty($headerFilename) && empty($this->headerFilename)) {
            // No header file name; set a default
            $this->headerFilename = $this->filename . '.dat';
        } elseif (!empty($headerFilename)) {
            $this->headerFilename = $headerFilename;
        }

        if (!file_exists($this->filename)) {
            throw new File_Fortune_Exception_BadFile(
                'Bad or missing fortune file (' . $this->filename . ')'
            );
        }

        if (false === ($file = fopen($this->filename, 'r'))) {
            throw new File_Fortune_Exception_BadFile(
                'Unable to open fortune file (' . $this->filename . ')'
            );
        }
        $this->file =& $file;

        try {
            $this->_readHeader();
        } catch (File_Fortune_Exception $e) {
            throw $e;
        }

        return true;
    }

    /**
     * Close fortune file
     *
     * Closes the fortune file if it's open, and unsets all header-related
     * properties (except {@link $delim}); does nothing otherwise.
     *
     * @return true
     * @access public
     */
    public function close()
    {
        if (isset($this->file)) {
            unset(
                $this->flags, 
                $this->maxLength,
                $this->minLength,
                $this->numstr,
                $this->offsets,
                $this->version,
                $this->fortunes
            );

            fclose($this->file);
        }

        return true;
    }

    /**
     * Read a fortune database header file
     * 
     * Reads the header file associated with this fortune database.  The name of
     * the header file is determined by the {@link __construct() constructor}:
     * either it is based on the name of the data file, or supplied by the
     * caller.
     * 
     * If the header file does not exist, this function calls 
     * {@link _computeHeader()} automatically, which has the same effect as
     * reading the header from a file.
     * 
     * The header contains the following values, which are stored as attributes
     * of the {@link File_Fortune} object:
     * <ul>
     *     <li>{@link $version}; version number</li>
     *     <li>{@link $numstr}; number of strings (fortunes) in the file</li>
     *     <li>{@link $maxLength}; length of longest string in the file</li>
     *     <li>{@link $minLength}; length of shortest string in the file</li>
     *     <li>{@link $flags}; bit field for flags (see strfile(1) man
     *     page)</li>
     *     <li>{@link $delim}; character that delimits fortunes</li>
     * </ul>
     *
     * {@link $numstr} is available via the {@link numFortunes()} method; if
     * you're interested in the others, you'll have to go grubbing through the
     * File_Fortune object, e.g.:
     *
     * <code>
     * $fortune_file = new Fortune('fortunes');
     * $delim        = $fortune_file->delim;
     * $maxLength    = $fortune_file->maxLength;
     * </code>
     *
     * {@link _readHeader()} throws an exception if there are any problems
     * reading the header file, e.g. if it seems to be corrupt or truncated.
     *
     * @access protected
     * @return true
     * @throws File_Fortune_Exception_BadHeaderFile if unable to open header
     * file, unable to read header file, unable to get header information, or if
     * header information appears corrupted
     * @throws File_Fortune_Exception_BadFile if unable to open fortune file if
     * computing headers directly from fortune file is necessary
     */
    protected function _readHeader()
    {
        $filename = $this->headerFilename;
        if (!file_exists($filename) && file_exists($this->filename)) {
            try {
                return $this->_computeHeader();
            } catch (File_Fortune_Exception_BadFile $e) {
                return $e;
            }
        }
      
        if (false === ($hdrFile = fopen($filename, 'r'))) {
            throw new File_Fortune_Exception_BadHeaderFile(
                'Unable to open header file (' . $filename . ')'
            );
        }

        /* 
         * from the strfile(1) man page:
         *       unsigned long str_version;  // version number
         *       unsigned long str_numstr;   // # of strings in the file
         *       unsigned long str_longlen;  // length of longest string
         *       unsigned long str_shortlen; // shortest string length
         *       unsigned long str_flags;    // bit field for flags
         *       char str_delim;             // delimiting character
         * that 'char' is padded out to a full word, so the header is 24 bytes
         */

        if (false === ($header = fread($hdrFile, 24))) {
            throw new File_Fortune_Exception_BadHeaderFile(
                'Unable to read header file (' . $filename . ')'
            );
        }

        $values = unpack('N5nums/adelim/xxx', $header);
        if (empty($values) || (6 != count($values))) {
            throw new File_Fortune_Exception_BadHeaderFile(
                'Failed to load full header'
            );
        }

        $this->version    = $values['nums1'];
        $this->numstr     = $values['nums2'];
        $this->maxLength  = $values['nums3'];
        $this->minLength  = $values['nums4'];
        $this->flags      = $values['nums5'];
        $this->delim      = $values['delim'];

        $expectedOffsets = $this->numstr + 1;

        $amountData = 4 * $expectedOffsets;

        if (false === ($data = fread($hdrFile, $amountData))) {
            throw new File_Fortune_Exception_BadHeaderFile(
                'Failed to read offsets for all fortunes'
            );
        }
        $offsets = array_merge(unpack('N*', $data));
        if (count($offsets) != $expectedOffsets) {
            throw new File_Fortune_Exception_BadHeaderFile(
                'Expected offsets does not match actual offsets'
            );
        }
        $this->offsets = $offsets;

        fclose($hdrFile);
        return true;
    }


    /**
     * Compute fortune file header information
     *
     * Reads the contents of the fortune file and computes the header
     * information that would normally be found in a header (.dat) file and read
     * by {@link _readHeader()}.  This is useful if you maintain a file of
     * fortunes by hand and do not have the corresponding data file. (If this is
     * the case, you may also want to look at 
     * {@link File_Fortune_Writer::write()}.
     * 
     * An optional delimiter argument may be passed to this function; if
     * present, that delimiter will be used to separate entries in the fortune
     * file.  If not provided, a percent sign ("%") will be used (this is the
     * standard fortune file delimiter).
     *
     * Additionally, the {@link $noHeader} property will be set to true.
     *
     * <b>NOTE:</b> It is most efficient to use fortune files that have header
     * files. In order to get offsets and fortunes, this method actually must
     * read the entire fortune file, and stores all fortunes in the 
     * {@link $fortunes} array.
     *
     * @access protected
     * @param string $delim Defaults to '%'
     * @return true
     * @throws File_Fortune_Exception_BadFile if unable to open fortune file
     */
    protected function _computeHeader($delim = '%')
    {
        $filename = $this->filename;
        if (false === ($contents = file_get_contents($filename))) {
            throw new File_Fortune_Exception_BadFile(
                'Unable to open fortune file (' . $filename . ') to compute headers'
            );
        }

        // Get all fortunes
        // Strip off final delimiter
        $contents = substr_replace($contents, '', strrpos($contents, "$delim\n"));
        $fortunes = explode("$delim\n", $contents); // explode into array

        // Get offsets, min, max, etc.
        $offsets   = array(0);
        $numstr    = count($fortunes);
        $curOffset = 0;
        $delimLen  = strlen($delim);
        $min       = null;
        $max       = null;
        for ($i = 0; $i < $numstr; $i++) {
            $len        = strlen($fortunes[$i]);
            $curOffset += $len + $delimLen + 1;
            $offsets[]  = $curOffset;
            if (empty($min) || ($min > $len)) {
                $min = $len;
            } elseif (empty($max) || ($max < $len)) {
                $max = $len;
            }
        }

        // Set object properties
        $this->version    = 1;
        $this->numstr     = $numstr;
        $this->maxLength  = $max;
        $this->minLength  = $min;
        $this->flags      = 0;
        $this->delim      = $delim;
        $this->offsets    = $offsets;
        $this->fortunes   = $fortunes;
        $this->noHeader   = true;

        return true;
    }

    /**
     * Return number of fortunes
     *
     * Returns the number of fortunes found by {@link readHeader()}. If the
     * {@link $numstr} property has not been set, throws an exception.
     * 
     * @access public
     * @return int|false
     * @throws File_Fortune_Exception_HeaderNotRead if {@link $numstr} is empty.
     */
    public function num()
    {
        if (empty($this->numstr)) {
            throw new File_Fortune_Exception_HeaderNotRead(
                'Header not read'
            );
        }
        return $this->numstr;
    }

    /**
     * Return a fortune
     *
     * Reads string number $num from the open fortune file. $num is ordinal, ie.
     * it must be between 1 and {@link $numstr} (inclusive).  Throws an
     * exception if you haven't opened the file and read the header, or if $num
     * is out of range.  (Opening the file is pretty hard to screw up, since
     * it's taken care of for you by {@link open()}, and any real errors should
     * occur during that call.)  Returns the text of the fortune as a (possibly)
     * multiline string.
     *
     * In order to read a fortune, one of the following actions is taken:
     * <ul>
     *     <li>If {@link $fortunes} is populated, it grabs the fortune from
     *     there.</li>
     *     <li>Otherwise, it utilizes {@link _readFromFile()} to grab the
     *     fortune from the fortune file.</li>
     * </ul>
     *
     * @access public
     * @param int $num Index of fortune to return
     * @return string|false
     * @throws File_Fortune_Exception_NotOpen if the fortune file hasn't yet been
     * opened
     * @throws File_Fortune_Exception_InvalidCount if an invalid fortune number
     * is passed
     * @throws File_Fortune_Exception_NotOpen if fortune file is not open
     * @throws File_Fortune_Exception_InvalidID if ID is out of range
     * @throws File_Fortune_Exception_Offset if unable to seek to fortune offset
     */
    public function read($num = 1)
    {
        if (!empty($this->fortunes)
            && is_array($this->fortunes) 
            && isset($this->fortunes[$num - 1])) 
        {
            return $this->fortunes[$num - 1];
        }

        if (!empty($this->fortunes)
            && is_array($this->fortunes) 
            && !isset($this->fortunes[$num - 1])) 
        {
            throw new File_Fortune_Exception_InvalidCount(
                'Invalid fortune number (array)'
            );
        }

        if (empty($this->fortunes)) {
            try {
                return $this->_readFromFile($num);
            } catch (File_Fortune_Exception $e) {
                throw $e;
            }
        }

        throw new File_Fortune_Exception_NotOpen('Fortune file not open');
    }

    /**
     * Read a fortune from the file
     *
     * Reads a fortune directly from the file. If an error occurs, an exception
     * is thrown. Otherwise, on sucess, the fortune is returned.
     *
     * @access protected
     * @param int
     * @return string|false
     * @throws File_Fortune_Exception_NotOpen if fortune file is not open
     * @throws File_Fortune_Exception_InvalidID if ID is out of range
     * @throws File_Fortune_Exception_Offset if unable to seek to fortune offset
     */
    protected function _readFromFile($num) 
    {
        if (empty($this->file)) {
            throw new File_Fortune_Exception_NotOpen('Fortune file not open');
        }

        if (!isset($this->offsets[$num]) || !isset($this->offsets[$num - 1])) {
            throw new File_Fortune_Exception_InvalidID('Invalid fortune number');
        }

        $start  = $this->offsets[$num - 1];
        $end    = $this->offsets[$num];
        $length = $end - $start;

        // decrement length 2 bytes for most fortunes (to drop trailing "%\n"),
        // and none for the last one (keep trailing newline)
        $delimLength = strlen($this->delim) | 1;
        $length -= ($num == $this->numstr) ? 0 : ($delimLength + 1);

        if (-1 == fseek($this->file, $start)) {
            throw new File_Fortune_Exception_Offset(
                'Unable to seek to fortune offset'
            );
        }
        $fortune = fread($this->file, $length);

        return $fortune;
    }

    /**
     * Return a random fortune
     *
     * Picks a random fortune for you and reads and returns it with 
     * {@link read()}.  If the {@link $numstr} propery has not been set, throws
     * an exception.
     *
     * @access public
     * @return string|false
     * @throws File_Fortune_Exception_HeaderNotRead if header file hasn't been
     * read
     * @throws File_Fortune_Exception_NotOpen if the fortune file hasn't yet been
     * opened
     * @throws File_Fortune_Exception_InvalidCount if an invalid fortune number
     * is passed
     * @throws File_Fortune_Exception_NotOpen if fortune file is not open
     * @throws File_Fortune_Exception_InvalidID if ID is out of range
     * @throws File_Fortune_Exception_Offset if unable to seek to fortune offset
     */
    public function getRandom()
    {
        if (empty($this->numstr)) {
            throw new File_Fortune_Exception_HeaderNotRead(
                'Header file not read'
            );
        }

        $num = rand(1, $this->numstr);
        try {
            return $this->read($num);
        } catch (File_Fortune_Exception $e) {
            return $e;
        }
    }

    /**
     * Retrieve all fortunes in a file
     *
     * Retrieves all fortunes in a file, returning them as an array and placing
     * them in the {@link $fortunes} property. If the {@link $numstr} property
     * has not been set, or if an error occurs elsewhere along the chain, throws
     * an exception.
     *
     * @access public
     * @return array|false
     * @throws File_Fortune_Exception_HeaderNotRead if header file hasn't been
     * read
     * @throws File_Fortune_Exception_NotOpen if the fortune file hasn't yet been
     * opened
     * @throws File_Fortune_Exception_InvalidCount if an invalid fortune number
     * is passed
     * @throws File_Fortune_Exception_NotOpen if fortune file is not open
     * @throws File_Fortune_Exception_InvalidID if ID is out of range
     * @throws File_Fortune_Exception_Offset if unable to seek to fortune offset
     */
    public function getAll() 
    {
        if (!empty($this->fortunes) && is_array($this->fortunes)) {
            return $this->fortunes;
        }

        if (empty($this->numstr)) {
            throw new File_Fortune_Exception_HeaderNotRead('Header file not read');
        }

        $num      = $this->numstr;
        $fortunes = array();
        for ($i = 0; $i < $num; $i++) {
            try {
                $fortune = $this->read($i + 1);
            } catch (File_Fortune_Exception $e) {
                throw $e;
            }
            $fortunes[] = $fortune;
        }

        $this->fortunes = $fortunes;

        return $fortunes;
    }

    /**
     * Retrieve a random fortune from a list of fortune files
     *
     * Returns a random fortune from a list of fortune files. The list of files
     * may be either specified as individual arguments, as a single array
     * argument, or as a mixture of individual files and arrays of files.
     * Invalid arguments -- non-files -- will be skipped.
     *
     * @access public
     * @static
     * @return string
     * @throws File_Fortune_Exception_NoFilename if no filenames provided
     * @throws File_Fortune_Exception subclass based on result of {@link open()}
     * and {@link getRandom()} calls
     */
    public static function getRandomFromSet() 
    {
        $args  = func_get_args();
        $files = File_Fortune::_fileLoop($args);
        if (empty($files)) {
            throw new File_Fortune_Exception_NoFilename('No filenames found');
        }

        $file = $files[array_rand($files)];
        try {
            $ffile = new File_Fortune($file);
            return $ffile->getRandom();
        } catch (File_Fortune_Exception $e) {
            throw $e;
        }
    }

    /**
     * Loop through an array looking for files
     *
     * Loops through an array, searching for files. An array of unique files
     * found is returned.
     *
     * @static
     * @access private
     * @param array
     * @return array
     */
    public static function _fileLoop($array) 
    {
        $files = array();
        foreach ($array as $file) {
            if (is_string($file) && file_exists($file)) {
                $files[] = $file;
                continue;
            }
            if (is_array($file)) {
                $tmp = File_Fortune::_fileLoop($file);
                if (!empty($tmp)) {
                    $files = array_merge($files, $tmp);
                }
            }
        }

        $files = array_unique($files);
        return $files;
    }
}
?>
