<?php

/* vim: set expandtab tabstop=4 shiftwidth=4 softtabstop=4: */

/**
 * File_Fortune_Writer: Write and modify fortune files
 *
 * Write and modify fortune files in PHP. Inherits from {@link File_Fortune} for
 * read and header functions, and adds methods for writing fortune files. Uses
 * exception classes provided in {@link File_Fortune_Exception} for exception
 * handling.
 *
 * @package   File_Fortune
 * @author    Matthew Weier O'Phinney <mweierophinney@gmail.com>
 * @copyright 2005, Matthew Weier O'Phinney
 * @version   CVS: $Id:$
 */

/**
 * Require File_Fortune as the base class
 */
require_once 'File/Fortune.php';

/**
 * Utilize exceptions for error handling
 */
require_once 'File/Fortune/Exception.php';

/**
 * File_Fortune_Writer: Write and modify fortune files
 *
 * Write and modify fortune files in PHP. Inherits from {@link File_Fortune} for
 * read and header functions, and adds functions for writing fortune files.
 *
 * <b>DESCRIPTION</b>
 *
 * This package provides the ability to write fortune files. It can:
 * <ul>
 *     <li>{@link write() write} entire files, given an array of fortunes</li>
 *     <li>{@link add() add} a fortune to an existing file</li>
 *     <li>{@link update() update} an existing fortune in an existing file</li>
 *     <li>{@link delete() delete} an existing fortune from an existing
 *     file</li>
 * </ul>
 *
 * Whenever files are written, a binary header file is also written. This header
 * file, which by default is named after the fortune file with an extra '.dat'
 * extension, contains information on the number of fortunes in the fortune file
 * and the offset of each fortune within the file. Files written with
 * File_Fortune_Writer will be compatible with all newer versions of fortune and
 * fortune-mod.
 *
 * Older style fortune files that munged the header and data into a single file
 * are not currently supported, and may never be.
 *
 * The class inherits its constructor from {@link File_Fortune}.
 *
 * @package   File_Fortune
 * @version   Release: 0.9.0
 * @tutorial  File_Fortune/File_Fortune.cls
 */
class File_Fortune_Writer extends File_Fortune
{
    /**
     * Write a fortune file
     *
     * Writes a fortune file and the related header file.
     *
     * The $fortunes array should be an ordinal array of fortune strings.
     *
     * The file is written to {@link $filename}, and the header file to 
     * {@link $headerFilename}. If {@link $headerFilename} remains undefined,
     * the filename {@link $filename}.dat is used.
     *
     * On error, an exception is thrown.
     *
     * @access public
     * @param array $fortunes Array of fortunes
     * @return bool
     * @throws File_Fortune_Writer_Exception_FileOpen if unable to open file
     * @throws File_Fortune_Writer_Exception_FileLock if unable to obtain file
     * lock
     * @throws File_Fortune_Writer_Exception_FilePos if unable to obtain one or
     * more file offsets while writing file
     */
    public function write($fortunes) 
    {
        $this->close();

        if (false === ($fh = fopen($this->filename, 'w'))) {
            throw new File_Fortune_Writer_Exception_FileOpen(
                'Unable to open ' . $this->filename
            );
        }
        if (!flock($fh, LOCK_EX)) {
            throw new File_Fortune_Writer_Exception_FileLock(
                'Unable to obtain file lock'
            );
        }

        // Write file
        $offsets = array(0); // obtain offsets as we go
        $min     = null;     // For determining shortest fortune
        $max     = null;     // For determining longest fortune
        foreach ($fortunes as $fortune) {
            fwrite($fh, $fortune);
            fwrite($fh, $this->delim . "\n");
            if (false === ($pos = ftell($fh))) {
                throw new File_Fortune_Writer_Exception_FilePos(
                    'Unable to obtain file position'
                );
            }

            // Determine if this is the min or max length fortune
            $len     = strlen($fortune);
            if (empty($min) || ($len < $min)) { 
                $min = $len;
            } elseif (empty($max) || ($len > $max)) { 
                $max = $len;
            }

            // Add offset to list
            $offsets[] = $pos;
        }
        flock($fh, LOCK_UN);
        fclose($fh);


        // Create header block
        // Set version
        if (empty($this->version)) {
            $this->version = 1;
        } else {
            $this->version++;
        }

        // Set other header attributes
        $this->offsets   = $offsets;         // list of offsets
        $this->numstr    = count($fortunes); // number of fortunes
        $this->maxLength = $max;             // longest fortune
        $this->minLength = $min;             // shortest fortune
        if (empty($this->flags)) {
            $this->flags = 0;
        }

        // Write header file
        try {
            $e = $this->_writeHeader();
        } catch (File_Fortune_Writer_Exception $e) {
            throw $e;
        }

        // Reload fortune file
        try {
            return $this->open();
        } catch (File_Fortune_Exception $e) {
            throw $e;
        }
    }

    /**
     * Write the header file for a fortune file
     *
     * Writes a fortune header file to {@link $headerFilename}, using 
     * {@link $version}, {@link $numstr}, {@link $maxLength}, 
     * {@link $minLength}, {@link $flags}, {@link $delim}, and {@link $offsets}.
     *
     * If an error occurs, an exception is thrown.
     *
     * @access protected
     * @return true
     * @throws File_Fortune_Writer_Exception_FileOpen if unable to open header
     * file
     * @throws File_Fortune_Writer_Exception_FileLock if unable to obtain lock on
     * header file
     */
    protected function _writeHeader() 
    {
        if (empty($this->headerFilename)) {
            $this->headerFilename = $this->filename . '.dat';
        }

        $header = pack(
            'NNNNNaxxx', 
            $this->version,
            $this->numstr,
            $this->maxLength,
            $this->minLength,
            $this->flags,
            $this->delim
        );

        // Pack offsets
        $offsetBin = '';
        foreach ($this->offsets as $offset) {
            $offsetBin .= pack('N', $offset);
        }

        // Write header file
        if (false === ($fh = fopen($this->headerFilename, 'w'))) {
            throw new File_Fortune_Writer_Exception_FileOpen(
                'Unable to open header file ' . $this->headerFilename
            );
        }
        if (!flock($fh, LOCK_EX)) {
            throw new File_Fortune_Writer_Exception_FileLock(
                'Unable to obtain lock on header file'
            );
        }
        fwrite($fh, $header . $offsetBin);
        flock($fh, LOCK_UN);
        fclose($fh);

        return true;
    }

    /**
     * Update a single fortune in a file
     *
     * Updates a single fortune in a file, as denoted by $num, with $fortune.
     * If $num is greater than the number of fortunes currently in the file, the
     * fortune is added as the last fortune in the file.
     *
     * <b>Note:</b> Updating is an expensive operation. All offsets following
     * the updated fortune need to be recalculated -- which means that all
     * fortunes need to be loaded in memory first, then written to the file.
     * It's probably best to stick to {@link add() adding} fortunes or 
     * {@link read() reading} fortunes in most instances.
     *
     * On error, an exception is thrown.
     *
     * @access public
     * @param int
     * @param string
     * @return bool
     * @throws File_Fortune_Exception_HeaderNotRead if header file hasn't been
     * read
     * @throws File_Fortune_Exception_NotOpen if the fortune file hasn't yet been
     * opened
     * @throws File_Fortune_Exception_InvalidCount if an invalid fortune number
     * is passed
     * @throws File_Fortune_Exception_NotOpen if fortune file is not open
     * @throws File_Fortune_Exception_InvalidID if ID is out of range
     * @throws File_Fortune_Exception_Offset if unable to seek to fortune offset
     * @throws File_Fortune_Writer_Exception_FileOpen if unable to open file
     * @throws File_Fortune_Writer_Exception_FileLock if unable to obtain file
     * lock
     * @throws File_Fortune_Writer_Exception_FilePos if unable to obtain one or
     * more file offsets while writing file
     */
    public function update($num, $fortune)
    {
        // Make sure fortune has an EOL
        if (!preg_match("/\n$/s", $fortune)) $fortune .= "\n";

        try {
            $fortunes = $this->getAll();
        } catch (File_Fortune_Exception $e) {
            throw $e;
        }

        if (isset($fortunes[$num - 1])) {
            $fortunes[$num - 1] = $fortune;
        } else {
            $fortunes[] = $fortune;
        }

        try {
            return $this->write($fortunes);
        } catch (File_Fortune_Writer_Exception $e) {
            throw $e;
        }
    }

    /**
     * Add a new fortune
     *
     * Adds a new fortune to the file. In doing so, it simply appends the file
     * with a new fortune, and then rewrites the header file with new
     * information (new number of strings, new version, possibly new min/max,
     * new offset).
     *
     * On error, an exception is thrown.
     *
     * @access public
     * @param string $fortune
     * @return true
     * @throws File_Fortune_Writer_Exception_FileOpen if unable to open file
     * @throws File_Fortune_Writer_Exception_FileLock if unable to obtain file
     * lock
     * @throws File_Fortune_Writer_Exception_FilePos if unable to obtain one or
     * more file offsets while writing file
     * @throws File_Fortune_Exception_NoFilename if no filename passed and no
     * filename in object
     * @throws File_Fortune_Exception_BadFile if missing file or unable to read
     * from file
     * @throws File_Fortune_Exception_BadHeaderFile if unable to open header
     * file, unable to read header file, unable to get header information, or if
     * header information appears corrupted
     */
    public function add($fortune) 
    {
        // Make sure fortune has an EOL
        if (!preg_match("/\n$/s", $fortune)) $fortune .= "\n";

        // Get delimiter
        if (empty($this->delim)) {
            $this->delim = '%';
        }

        // If no header file, then simply write the file
        if (isset($this->noHeader)) {
            if (!isset($this->fortunes)) $this->fortunes = array();
            $this->fortunes[] = $fortune;
            try {
                return $this->write($this->fortunes);
            } catch (File_Fortune_Writer_Exception $e) {
                throw $e;
            }
        }

        if (false === ($fh = fopen($this->filename, 'a'))) {
            throw new File_Fortune_Writer_Exception_FileOpen(
                'Unable to open ' . $this->filename . ' for append'
            );
        }
        if (!flock($fh, LOCK_EX)) {
            throw new File_Fortune_Writer_Exception_FileLock(
                'Unable to obtain file lock'
            );
        }

        // Write fortune and delimiter
        fwrite($fh, $fortune);
        fwrite($fh, $this->delim . "\n");

        // Calculate new offset
        if (false === ($pos = ftell($fh))) {
            throw new File_Fortune_Writer_Exception_FilePos(
                'Unable to obtain file position'
            );
        }
        fclose($fh);
        // Add $pos to last offset
        $last = array_pop($this->offsets);
        $end  = $last + $pos;
        array_push($this->offsets, $last, $end);

        // Set header elements
        $this->version++;        // increment version

        // Determine if its a max/min
        $len = strlen($fortune);
        if ($len < $this->minLength) {
            $this->minLength = $len;
        } elseif ($len > $this->maxLength) {
            $this->maxLength = $len;
        }

        // Write header file
        try {
            $this->_writeHeader();
        } catch (File_Fortune_Writer_Exception $e) {
            throw $e;
        }

        // Reopen the file
        $this->close();
        try {
            return $this->open();
        } catch (File_Fortune_Exception $e) {
            throw $e;
        }
    }

    /**
     * Remove a fortune
     *
     * Removes a fortune from the file, as specified by $num. 
     *
     * <b>Note:</b> Like {@link update() updating}, deleting is an expensive
     * operation. All offsets following the deleted fortune need to be
     * recalculated -- which means that all fortunes need to be loaded in memory
     * first, then re-written to the file.  It's probably best to stick to
     * {@link add() adding} fortunes or {@link read() reading} fortunes in most
     * instances.
     *
     * On error, an exception is thrown.
     *
     * @access public
     * @return true
     * @throws File_Fortune_Exception_HeaderNotRead if header file hasn't been
     * read
     * @throws File_Fortune_Exception_NotOpen if the fortune file hasn't yet been
     * opened
     * @throws File_Fortune_Exception_InvalidCount if an invalid fortune number
     * is passed
     * @throws File_Fortune_Exception_NotOpen if fortune file is not open
     * @throws File_Fortune_Exception_InvalidID if ID is out of range
     * @throws File_Fortune_Exception_Offset if unable to seek to fortune offset
     * @throws File_Fortune_Writer_Exception_FileOpen if unable to open file
     * @throws File_Fortune_Writer_Exception_FileLock if unable to obtain file
     * lock
     * @throws File_Fortune_Writer_Exception_FilePos if unable to obtain one or
     * more file offsets while writing file
     */
    public function delete($num) 
    {
        try {
            $fortunes = $this->getAll();
        } catch (File_Fortune_Exception $e) {
            throw $e;
        }

        if (isset($fortunes[$num - 1])) {
            unset($fortunes[$num - 1]);
            try {
                $this->write($fortunes);
            } catch (File_Fortune_Writer_Exception $e) {
                throw $e;
            }
        }

        return true;
    }
}

?>
